@@ -25,40 +25,49 @@
 # ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 # POSSIBILITY OF SUCH DAMAGE.
 
+from __future__ import annotations
 
-from yubikit.core import Tlv, BadResponseError, NotSupportedError
-from yubikit.core.smartcard import ApduError, SW
-from yubikit.piv import (
-    PivSession,
-    SLOT,
-    OBJECT_ID,
-    KEY_TYPE,
-    MANAGEMENT_KEY_TYPE,
-    ALGORITHM,
-    TAG_LRC,
-    SlotMetadata,
-    FascN,
-    Chuid,
-)
-from .util import display_serial
+import logging
+import os
+import re
+import struct
+import warnings
+from datetime import date, datetime
+from typing import TYPE_CHECKING, Any, Mapping, TypeAlias, cast
+from uuid import uuid4
 
 from cryptography import x509
 from cryptography.exceptions import InvalidSignature
-from cryptography.hazmat.primitives import hashes
-from cryptography.hazmat.primitives.asymmetric import rsa, ec, padding, ed25519, x25519
-from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat
-from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
 from cryptography.hazmat.backends import default_backend
+from cryptography.hazmat.primitives import hashes
+from cryptography.hazmat.primitives.asymmetric import ec, ed25519, padding, rsa, x25519
+from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
+from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat
 from cryptography.x509.oid import NameOID
-from datetime import datetime, date
-from uuid import uuid4
-import logging
-import struct
-import os
-import re
 
-from typing import Union, Mapping, Optional, List, Dict, Type, Any, cast
+from yubikit.core import BadResponseError, NotSupportedError, Tlv
+from yubikit.core.smartcard import SW, ApduError
+from yubikit.piv import (
+    ALGORITHM,
+    KEY_TYPE,
+    MANAGEMENT_KEY_TYPE,
+    OBJECT_ID,
+    SLOT,
+    TAG_LRC,
+    Chuid,
+    FascN,
+    PivSession,
+    SlotMetadata,
+)
 
+from .util import display_serial
+
+if TYPE_CHECKING:
+    # These types arent't available on cryptography <40.
+    from cryptography.hazmat.primitives.asymmetric.types import (
+        CertificatePublicKeyTypes,
+        PublicKeyTypes,
+    )
 
 logger = logging.getLogger(__name__)
 
@@ -83,7 +92,7 @@ _NAME_ATTRIBUTES = {
 _ESCAPED = "\\\"+,'<> #="
 
 
-def _parse(value: str) -> List[List[str]]:
+def _parse(value: str) -> list[list[str]]:
     remaining = list(value)
     name = []
     entry = []
@@ -121,15 +130,29 @@ def _parse(value: str) -> List[List[str]]:
 _DOTTED_STRING_RE = re.compile(r"\d(\.\d+)+")
 
 
+# TODO: Require cryptography >= 37 and remove this function
 def parse_rfc4514_string(value: str) -> x509.Name:
     """Parse an RFC 4514 string into a x509.Name.
 
     See: https://tools.ietf.org/html/rfc4514.html
 
     :param value: An RFC 4514 string.
+    :deprecated: Use x509.Name.from_rfc4514_string() instead.
     """
+    warnings.warn(
+        "Deprecated: use x509.Name.from_rfc4514_string() instead.",
+        DeprecationWarning,
+    )
+    try:
+        return x509.Name.from_rfc4514_string(value)
+    except AttributeError:
+        # cryptography < 37, use fallback implementation
+        return _parse_rfc4514_string(value)
+
+
+def _parse_rfc4514_string(value: str) -> x509.Name:
     name = _parse(value)
-    attributes: List[x509.RelativeDistinguishedName] = []
+    attributes: list[x509.RelativeDistinguishedName] = []
     for entry in name:
         parts = []
         for part in entry:
@@ -171,7 +194,7 @@ def derive_management_key(pin: str, salt: bytes) -> bytes:
     :param pin: The PIN.
     :param salt: The salt.
     """
-    kdf = PBKDF2HMAC(hashes.SHA1(), 24, salt, 10000, default_backend())  # nosec
+    kdf = PBKDF2HMAC(hashes.SHA1(), 24, salt, 10000, default_backend())  # noqa: S303
     return kdf.derive(pin.encode("utf-8"))
 
 
@@ -282,6 +305,10 @@ def get_pivman_protected_data(session: PivSession) -> PivmanProtectedData:
             logger.debug("No data, initializing blank")
             return PivmanProtectedData()
         raise
+    except Exception:
+        raise ValueError(
+            f"Invalid data in protected slot ({hex(OBJECT_ID_PIVMAN_PROTECTED_DATA)})"
+        )
 
 
 def pivman_set_mgm_key(
@@ -387,7 +414,7 @@ def pivman_set_pin_attempts(
         session.put_object(OBJECT_ID_PIVMAN_DATA, pivman.get_bytes())
 
 
-def list_certificates(session: PivSession) -> Mapping[SLOT, Optional[x509.Certificate]]:
+def list_certificates(session: PivSession) -> Mapping[SLOT, x509.Certificate | None]:
     """Read out and parse stored certificates.
 
     Only certificates which are successfully parsed are returned.
@@ -395,7 +422,8 @@ def list_certificates(session: PivSession) -> Mapping[SLOT, Optional[x509.Certif
     :param session: The PIV session.
     """
     certs = {}
-    for slot in set(SLOT) - {SLOT.ATTESTATION}:
+    slots: list[SLOT] = [s for s in SLOT if s != SLOT.ATTESTATION]
+    for slot in slots:
         try:
             certs[slot] = session.get_certificate(slot)
         except ApduError:
@@ -408,7 +436,8 @@ def list_certificates(session: PivSession) -> Mapping[SLOT, Optional[x509.Certif
 
 def _list_keys(session: PivSession) -> Mapping[SLOT, SlotMetadata]:
     keys = {}
-    for slot in set(SLOT) - {SLOT.ATTESTATION}:
+    slots: list[SLOT] = [s for s in SLOT if s != SLOT.ATTESTATION]
+    for slot in slots:
         try:
             keys[slot] = session.get_slot_metadata(slot)
         except ApduError as e:
@@ -420,7 +449,7 @@ def _list_keys(session: PivSession) -> Mapping[SLOT, SlotMetadata]:
 def check_key(
     session: PivSession,
     slot: SLOT,
-    public_key: Union[rsa.RSAPublicKey, ec.EllipticCurvePublicKey],
+    public_key: PublicKeyTypes,
 ) -> bool:
     """Check that a given public key corresponds to the private key in a slot.
 
@@ -455,7 +484,7 @@ def check_key(
         elif isinstance(public_key, ec.EllipticCurvePublicKey):
             public_key.verify(test_sig, test_data, ec.ECDSA(hashes.SHA256()))
         else:
-            raise ValueError("Unknown key type: " + type(public_key))
+            raise ValueError(f"Unknown key type: {type(public_key)}")
         return True
 
     except ApduError as e:
@@ -509,10 +538,10 @@ def get_piv_info(session: PivSession):
     :param session: The PIV session.
     """
     pivman = get_pivman_data(session)
-    info: Dict[str, Any] = {
+    info: dict[str, Any] = {
         "PIV version": session.version,
     }
-    lines: List[Any] = [info]
+    lines: list[Any] = [info]
 
     try:
         pin_data = session.get_pin_metadata()
@@ -528,9 +557,9 @@ def get_piv_info(session: PivSession):
     try:  # Bio metadata
         bio = session.get_bio_metadata()
         if bio.configured:
-            info[
-                "Biometrics"
-            ] = f"Configured, {bio.attempts_remaining} attempts remaining"
+            info["Biometrics"] = (
+                f"Configured, {bio.attempts_remaining} attempts remaining"
+            )
         else:
             info["Biometrics"] = "Not configured"
     except NotSupportedError:
@@ -564,7 +593,7 @@ def get_piv_info(session: PivSession):
     if pivman.has_stored_key:
         lines.append("Management key is stored on the YubiKey, protected by PIN.")
 
-    objects: Dict[str, Any] = {}
+    objects: dict[str, Any] = {}
     lines.append(objects)
     try:
         objects["CHUID"] = session.get_object(OBJECT_ID.CHUID)
@@ -583,11 +612,10 @@ def get_piv_info(session: PivSession):
         keys = _list_keys(session)
     except NotSupportedError:
         keys = {}
-    for slot in set(SLOT) - {SLOT.ATTESTATION}:
-        if slot not in keys and slot not in certs:
-            continue
 
-        cert_data: Dict[str, Any] = {}
+    slots = [s for s in SLOT if s in keys or s in certs]
+    for slot in slots:
+        cert_data: dict[str, Any] = {}
         objects[f"Slot {slot}"] = cert_data
         if slot in keys:
             cert_data["Private key type"] = keys[slot].key_type
@@ -612,7 +640,7 @@ def get_piv_info(session: PivSession):
             serial = cert.serial_number
             try:
                 try:  # Prefer timezone-aware variant (cryptography >= 42)
-                    not_before: Optional[datetime] = cert.not_valid_before_utc
+                    not_before: datetime | None = cert.not_valid_before_utc
                 except AttributeError:
                     not_before = cert.not_valid_before
             except ValueError:
@@ -620,7 +648,7 @@ def get_piv_info(session: PivSession):
                 not_before = None
             try:
                 try:  # Prefer timezone-aware variant (cryptography >= 42)
-                    not_after: Optional[datetime] = cert.not_valid_after_utc
+                    not_after: datetime | None = cert.not_valid_after_utc
                 except AttributeError:
                     not_after = cert.not_valid_after
             except ValueError:
@@ -643,16 +671,16 @@ def get_piv_info(session: PivSession):
     return lines
 
 
-_AllowedHashTypes = Union[
-    hashes.SHA224,
-    hashes.SHA256,
-    hashes.SHA384,
-    hashes.SHA512,
-    hashes.SHA3_224,
-    hashes.SHA3_256,
-    hashes.SHA3_384,
-    hashes.SHA3_512,
-]
+_AllowedHashTypes: TypeAlias = (
+    hashes.SHA224
+    | hashes.SHA256
+    | hashes.SHA384
+    | hashes.SHA512
+    | hashes.SHA3_224
+    | hashes.SHA3_256
+    | hashes.SHA3_384
+    | hashes.SHA3_512
+)
 
 
 def _hash(key_type, hash_algorithm):
@@ -666,7 +694,7 @@ def sign_certificate_builder(
     slot: SLOT,
     key_type: KEY_TYPE,
     builder: x509.CertificateBuilder,
-    hash_algorithm: Type[_AllowedHashTypes] = hashes.SHA256,
+    hash_algorithm: type[_AllowedHashTypes] = hashes.SHA256,
 ) -> x509.Certificate:
     """Sign a Certificate.
 
@@ -678,6 +706,7 @@ def sign_certificate_builder(
     """
     logger.debug("Signing a certificate")
     dummy_key = _dummy_key(key_type)
+    assert not isinstance(dummy_key, x25519.X25519PrivateKey)  # noqa: S101
     cert = builder.sign(dummy_key, _hash(key_type, hash_algorithm), default_backend())
 
     sig = session.sign(
@@ -700,9 +729,9 @@ def sign_certificate_builder(
 def sign_csr_builder(
     session: PivSession,
     slot: SLOT,
-    public_key: Union[rsa.RSAPublicKey, ec.EllipticCurvePublicKey],
+    public_key: PublicKeyTypes,
     builder: x509.CertificateSigningRequestBuilder,
-    hash_algorithm: Type[_AllowedHashTypes] = hashes.SHA256,
+    hash_algorithm: type[_AllowedHashTypes] = hashes.SHA256,
 ) -> x509.CertificateSigningRequest:
     """Sign a CSR.
 
@@ -716,6 +745,7 @@ def sign_csr_builder(
     logger.debug("Signing a CSR")
     key_type = KEY_TYPE.from_public_key(public_key)
     dummy_key = _dummy_key(key_type)
+    assert not isinstance(dummy_key, x25519.X25519PrivateKey)  # noqa: S101
 
     csr = builder.sign(dummy_key, _hash(key_type, hash_algorithm), default_backend())
     seq = Tlv.parse_list(Tlv.unpack(0x30, csr.public_bytes(Encoding.DER)))
@@ -749,11 +779,11 @@ def sign_csr_builder(
 def generate_self_signed_certificate(
     session: PivSession,
     slot: SLOT,
-    public_key: Union[rsa.RSAPublicKey, ec.EllipticCurvePublicKey],
+    public_key: PublicKeyTypes,
     subject_str: str,
     valid_from: datetime,
     valid_to: datetime,
-    hash_algorithm: Type[_AllowedHashTypes] = hashes.SHA256,
+    hash_algorithm: type[_AllowedHashTypes] = hashes.SHA256,
 ) -> x509.Certificate:
     """Generate a self-signed certificate using a private key in a slot.
 
@@ -767,6 +797,8 @@ def generate_self_signed_certificate(
     """
     logger.debug("Generating a self-signed certificate")
     key_type = KEY_TYPE.from_public_key(public_key)
+    if TYPE_CHECKING:
+        public_key = cast(CertificatePublicKeyTypes, public_key)
 
     subject = parse_rfc4514_string(subject_str)
     builder = (
@@ -785,9 +817,9 @@ def generate_self_signed_certificate(
 def generate_csr(
     session: PivSession,
     slot: SLOT,
-    public_key: Union[rsa.RSAPublicKey, ec.EllipticCurvePublicKey],
+    public_key: PublicKeyTypes,
     subject_str: str,
-    hash_algorithm: Type[_AllowedHashTypes] = hashes.SHA256,
+    hash_algorithm: type[_AllowedHashTypes] = hashes.SHA256,
 ) -> x509.CertificateSigningRequest:
     """Generate a CSR using a private key in a slot.
 
