@@ -25,15 +25,19 @@
 # ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 # POSSIBILITY OF SUCH DAMAGE.
 
+from __future__ import annotations
+
 import gzip
 import logging
 import os
 import re
 import warnings
+import zlib
+from base64 import b64decode
 from dataclasses import astuple, dataclass
 from datetime import date
 from enum import Enum, IntEnum, unique
-from typing import Optional, Union, cast, overload
+from typing import TYPE_CHECKING, TypeAlias, overload
 
 from cryptography import x509
 from cryptography.hazmat.backends import default_backend
@@ -70,21 +74,26 @@ from .core.smartcard import (
     SmartCardProtocol,
 )
 
+if TYPE_CHECKING:
+    # This type isn't available on cryptography <40.
+    from cryptography.hazmat.primitives.asymmetric.types import PublicKeyTypes
+
+
 logger = logging.getLogger(__name__)
 
 
-PublicKey = Union[
-    rsa.RSAPublicKey,
-    ec.EllipticCurvePublicKey,
-    ed25519.Ed25519PublicKey,
-    x25519.X25519PublicKey,
-]
-PrivateKey = Union[
-    rsa.RSAPrivateKeyWithSerialization,
-    ec.EllipticCurvePrivateKeyWithSerialization,
-    ed25519.Ed25519PrivateKey,
-    x25519.X25519PrivateKey,
-]
+PublicKey: TypeAlias = (
+    rsa.RSAPublicKey
+    | ec.EllipticCurvePublicKey
+    | ed25519.Ed25519PublicKey
+    | x25519.X25519PublicKey
+)
+PrivateKey: TypeAlias = (
+    rsa.RSAPrivateKeyWithSerialization
+    | ec.EllipticCurvePrivateKeyWithSerialization
+    | ed25519.Ed25519PrivateKey
+    | x25519.X25519PrivateKey
+)
 
 
 @unique
@@ -121,7 +130,7 @@ class KEY_TYPE(IntEnum):
         raise ValueError("No bit_len")
 
     @classmethod
-    def from_public_key(cls, key: PublicKey) -> "KEY_TYPE":
+    def from_public_key(cls, key: PublicKeyTypes) -> KEY_TYPE:
         if isinstance(key, rsa.RSAPublicKey):
             try:
                 return getattr(cls, "RSA%d" % key.key_size)
@@ -253,7 +262,7 @@ class OBJECT_ID(IntEnum):
     ATTESTATION = 0x5FFF01
 
     @classmethod
-    def from_slot(cls, slot: SLOT) -> "OBJECT_ID":
+    def from_slot(cls, slot: SLOT) -> OBJECT_ID:
         return getattr(cls, SLOT(slot).name)
 
 
@@ -295,6 +304,7 @@ INS_GET_DATA = 0xCB
 INS_PUT_DATA = 0xDB
 INS_MOVE_KEY = 0xF6
 INS_GET_METADATA = 0xF7
+INS_GET_SERIAL = 0xF8
 INS_ATTEST = 0xF9
 INS_SET_PIN_RETRIES = 0xFA
 INS_RESET = 0xFB
@@ -397,6 +407,12 @@ def _bcd(val, ln=1):
     return bits if ln == 1 else _bcd(val // 10, ln - 1) + bits
 
 
+def _dump_tlv_dict(values: dict[int, bytes | None]) -> bytes:
+    return b"".join(
+        Tlv(tag, value) for tag, value in values.items() if value is not None
+    )
+
+
 BCD_SS = "11010"
 BCD_FS = "10110"
 BCD_ES = "11111"
@@ -453,7 +469,7 @@ class FascN:
         return int2bytes(int(bs, 2) << 5 | lrc)
 
     @classmethod
-    def from_bytes(cls, value: bytes) -> "FascN":
+    def from_bytes(cls, value: bytes) -> FascN:
         bs = f"{bytes2int(value):0200b}"
         ds = [int(bs[i : i + 4][::-1], 2) for i in range(0, 200, 5)]
         args = (
@@ -467,53 +483,46 @@ class FascN:
         return "[%04d-%04d-%06d-%d-%d-%010d%d%04d%d]" % astuple(self)
 
 
-# From Python 3.10 we can use kw_only instead
-_chuid_no_value = object()
-
-
-@dataclass
+@dataclass(kw_only=True)
 class Chuid:
-    buffer_length: Optional[int] = None
-    fasc_n: FascN = cast(FascN, _chuid_no_value)
-    agency_code: Optional[bytes] = None
-    organizational_identifier: Optional[bytes] = None
-    duns: Optional[bytes] = None
-    guid: bytes = cast(bytes, _chuid_no_value)
-    expiration_date: date = cast(date, _chuid_no_value)
-    authentication_key_map: Optional[bytes] = None
-    asymmetric_signature: bytes = cast(bytes, _chuid_no_value)
-    lrc: Optional[int] = None
+    buffer_length: int | None = None
+    fasc_n: FascN
+    agency_code: bytes | None = None
+    organizational_identifier: bytes | None = None
+    duns: bytes | None = None
+    guid: bytes
+    expiration_date: date
+    authentication_key_map: bytes | None = None
+    asymmetric_signature: bytes
+    lrc: int | None = None
 
-    def __post_init__(self):
-        if _chuid_no_value in (
-            self.fasc_n,
-            self.guid,
-            self.expiration_date,
-            self.asymmetric_signature,
-        ):
-            raise ValueError("Missing required field(s)")
+    def _get_bytes(self, include_signature: bool = True) -> bytes:
+        return _dump_tlv_dict(
+            {
+                0xEE: int2bytes(self.buffer_length)
+                if self.buffer_length is not None
+                else None,
+                0x30: bytes(self.fasc_n),
+                0x31: self.agency_code,
+                0x32: self.organizational_identifier,
+                0x33: self.duns,
+                0x34: self.guid,
+                0x35: self.expiration_date.isoformat().replace("-", "").encode(),
+                0x3D: self.authentication_key_map,
+                0x3E: self.asymmetric_signature if include_signature else None,
+                TAG_LRC: bytes([self.lrc]) if self.lrc is not None else b"",
+            }
+        )
+
+    @property
+    def tbs_bytes(self) -> bytes:
+        return self._get_bytes(include_signature=False)
 
     def __bytes__(self):
-        bs = b""
-        if self.buffer_length is not None:
-            bs += Tlv(0xEE, int2bytes(self.buffer_length))
-        bs += Tlv(0x30, bytes(self.fasc_n))
-        if self.agency_code is not None:
-            bs += Tlv(0x31, self.agency_code)
-        if self.organizational_identifier is not None:
-            bs += Tlv(0x32, self.organizational_identifier)
-        if self.duns is not None:
-            bs += Tlv(0x33, self.duns)
-        bs += Tlv(0x34, self.guid)
-        bs += Tlv(0x35, self.expiration_date.isoformat().replace("-", "").encode())
-        if self.authentication_key_map is not None:
-            bs += Tlv(0x3D, self.authentication_key_map)
-        bs += Tlv(0x3E, self.asymmetric_signature)
-        bs += Tlv(TAG_LRC, bytes([self.lrc]) if self.lrc is not None else b"")
-        return bs
+        return self._get_bytes()
 
     @classmethod
-    def from_bytes(cls, value: bytes) -> "Chuid":
+    def from_bytes(cls, value: bytes) -> Chuid:
         data = Tlv.parse_dict(value)
         buffer_length = data.get(0xEE)
         lrc = data.get(TAG_LRC)
@@ -648,13 +657,64 @@ def _parse_device_public_key(key_type, encoded):
         return ec.EllipticCurvePublicKey.from_encoded_point(curve(), data[0x86])
 
 
+# This data embeds pre-computed structures used by the zlib deflate algorithm
+# It's purpose is to avoid to store the full dictionary in the compressed
+# stream, thus reducing the size of the certificate
+# Source: https://datatracker.ietf.org/doc/html/draft-pritikin-comp-x509-00#appendix-A
+_cxf_dictionary = b64decode("""
+MIIBOTCCASOgAwIBAgIBATANBgkqhkiG9w0BAQUFADAZMRcwFQYDVQQDEw5odHRw
+Oi8vd3d3LmNvbTAeFw0xMDA1MTExOTEzMDNaFw0xMTA1MTExOTEzMDNaMF8xEDAO
+BgkqhkiG9w0BCQEWAUAxCjAIBgNVBAMTASAxCzAJBgNVBAYTAlVTMQswCQYDVQQI
+EwJXSTELMAkGA1UEChMCb24xDDAKBgNVBAsTA291bjEKMAgGA1UEBRMBIDAfMA0G
+CSqGSIb3DQEBAQUAAw4AMAsCBG6G5ZUCAwEAAaNNMEswCQYDVR0TBAIwADAdBgNV
+HQ4EFgQUHSkK6busCxxK6PKpBlL9q8K1mcQwHwYDVR0jBBgwFoAUn7r/DVMuEpK9
+Rxq3nyiLml10+nQwDQYJKoZIhvcNAQEFBQADAQA=
+""")
+
+
+def decompress_certificate(cert_data: bytes) -> bytes:
+    """
+    Decompress a compressed certificate using various methods.
+    """
+    logger.debug("Certificate is compressed, decompressing...")
+
+    # CXF Deflate Method (as used by Pointsharp Net iD)
+    if cert_data[:2] == b"\01\00":
+        expected_length = int.from_bytes(cert_data[2:4], "little")
+        decompressor = zlib.decompressobj(wbits=zlib.MAX_WBITS, zdict=_cxf_dictionary)
+        try:
+            cert_data = decompressor.decompress(cert_data[4:])
+            if len(cert_data) != expected_length:
+                logger.error(
+                    "Unexpected decompressed length, expected %d, got %d",
+                    expected_length,
+                    len(cert_data),
+                )
+                raise ValueError("Decompressed length does not match expected length")
+
+            logger.debug("Decompressed certificate with CXF deflate format")
+            return cert_data
+        except (zlib.error, ValueError):
+            logger.warning("Failed to decompress with CXF format")
+
+    # Gzip Method (default)
+    try:
+        cert_data = gzip.decompress(cert_data)
+        logger.debug("Decompressed certificate with basic gzip format")
+        return cert_data
+    except (zlib.error, gzip.BadGzipFile):
+        logger.warning("Failed to decompressed with basic gzip format")
+
+    raise BadResponseError("Failed to decompress certificate")
+
+
 class PivSession:
     """A session with the PIV application."""
 
     def __init__(
         self,
         connection: SmartCardConnection,
-        scp_key_params: Optional[ScpKeyParams] = None,
+        scp_key_params: ScpKeyParams | None = None,
     ):
         self.protocol = SmartCardProtocol(connection)
         self.protocol.select(AID.PIV)
@@ -740,10 +800,18 @@ class PivSession:
 
         logger.info("PIV application data reset performed")
 
+    def get_serial(self) -> int:
+        """Get the serial number of the YubiKey."""
+        logger.debug("Getting serial number")
+        require_version(self.version, (5, 0, 3))
+        response = self.protocol.send_apdu(0, INS_GET_SERIAL, 0, 0)
+        return bytes2int(response)
+
     @overload
     def authenticate(self, management_key: bytes) -> None: ...
 
     @overload
+    # TODO: remove in 6.0
     def authenticate(
         self, key_type: MANAGEMENT_KEY_TYPE, management_key: bytes
     ) -> None: ...
@@ -786,7 +854,7 @@ class PivSession:
 
         backend = default_backend()
         cipher_key = _parse_management_key(key_type, management_key)
-        cipher = Cipher(cipher_key, modes.ECB(), backend)  # nosec
+        cipher = Cipher(cipher_key, modes.ECB(), backend)  # noqa: S305
         decryptor = cipher.decryptor()
         decrypted = decryptor.update(witness) + decryptor.finalize()
 
@@ -839,6 +907,9 @@ class PivSession:
     def verify_pin(self, pin: str) -> None:
         """Verify the user by PIN.
 
+        NOTE: InvalidPinError raised from this method will cap remaining_attempts
+        at 15, even if the true number of remaining attempts is higher.
+
         :param pin: The PIN.
         """
         logger.debug("Verifying PIN")
@@ -854,7 +925,7 @@ class PivSession:
 
     def verify_uv(
         self, temporary_pin: bool = False, check_only: bool = False
-    ) -> Optional[bytes]:
+    ) -> bytes | None:
         """Verify the user by fingerprint (YubiKey Bio only).
 
         Fingerprint verification will allow usage of private keys which have a PIN
@@ -938,6 +1009,9 @@ class PivSession:
     def change_pin(self, old_pin: str, new_pin: str) -> None:
         """Change the PIN.
 
+        NOTE: InvalidPinError raised from this method will cap remaining_attempts
+        at 15, even if the true number of remaining attempts is higher.
+
         :param old_pin: The current PIN.
         :param new_pin: The new PIN.
         """
@@ -948,6 +1022,9 @@ class PivSession:
     def change_puk(self, old_puk: str, new_puk: str) -> None:
         """Change the PUK.
 
+        NOTE: InvalidPinError raised from this method will cap remaining_attempts
+        at 15, even if the true number of remaining attempts is higher.
+
         :param old_puk: The current PUK.
         :param new_puk: The new PUK.
         """
@@ -963,6 +1040,9 @@ class PivSession:
     def unblock_pin(self, puk: str, new_pin: str) -> None:
         """Reset PIN with PUK.
 
+        NOTE: InvalidPinError raised from this method will cap remaining_attempts
+        at 15, even if the true number of remaining attempts is higher.
+
         :param puk: The PUK.
         :param new_pin: The new PIN.
         """
@@ -1071,8 +1151,8 @@ class PivSession:
         slot: SLOT,
         key_type: KEY_TYPE,
         message: bytes,
-        hash_algorithm: hashes.HashAlgorithm,
-        padding: Optional[AsymmetricPadding] = None,
+        hash_algorithm: hashes.HashAlgorithm | None,
+        padding: AsymmetricPadding | None = None,
     ) -> bytes:
         """Sign message with key.
 
@@ -1110,8 +1190,8 @@ class PivSession:
         except AttributeError:
             raise ValueError("Invalid length of ciphertext")
         logger.debug(
-            f"Decrypting data with key in slot {slot} of type {key_type} using ",
-            f"padding={padding}",
+            f"Decrypting data with key in slot {slot} of type {key_type} using "
+            f"padding={padding}"
         )
         padded = self._use_private_key(slot, key_type, cipher_text, False)
         return _unpad_message(padded, padding)
@@ -1119,9 +1199,9 @@ class PivSession:
     def calculate_secret(
         self,
         slot: SLOT,
-        peer_public_key: Union[
-            ec.EllipticCurvePublicKeyWithSerialization, x25519.X25519PublicKey
-        ],
+        peer_public_key: (
+            ec.EllipticCurvePublicKeyWithSerialization | x25519.X25519PublicKey
+        ),
     ) -> bytes:
         """Calculate shared secret using ECDH.
 
@@ -1172,7 +1252,7 @@ class PivSession:
         except ValueError as e:
             raise BadResponseError("Malformed object data", e)
 
-    def put_object(self, object_id: int, data: Optional[bytes] = None) -> None:
+    def put_object(self, object_id: int, data: bytes | None = None) -> None:
         """Write data to PIV object.
 
         Requires authentication with management key.
@@ -1205,11 +1285,7 @@ class PivSession:
 
         if cert_info == 1:
             # Compressed certificate
-            logger.debug("Certificate is compressed, decompressing...")
-            try:
-                cert_data = gzip.decompress(cert_data)
-            except gzip.BadGzipFile:
-                raise BadResponseError("Unable to decompress certificate")
+            cert_data = decompress_certificate(cert_data)
         elif cert_info != 0:
             raise NotSupportedError("Unsupported value in CertInfo")
 
@@ -1279,9 +1355,9 @@ class PivSession:
         self.check_key_support(key_type, pin_policy, touch_policy, False)
         ln = key_type.bit_len // 8
         if key_type.algorithm == ALGORITHM.RSA:
-            assert isinstance(private_key, rsa.RSAPrivateKey)  # nosec
+            assert isinstance(private_key, rsa.RSAPrivateKey)  # noqa: S101
             numbers = private_key.private_numbers()
-            numbers = cast(rsa.RSAPrivateNumbers, numbers)
+            numbers = numbers
             if numbers.public_numbers.e != 65537:
                 raise NotSupportedError("RSA exponent must be 65537")
             ln //= 2
@@ -1300,9 +1376,9 @@ class PivSession:
                 ),
             )
         else:
-            assert isinstance(private_key, ec.EllipticCurvePrivateKey)  # nosec
+            assert isinstance(private_key, ec.EllipticCurvePrivateKey)  # noqa: S101
             numbers = private_key.private_numbers()
-            numbers = cast(ec.EllipticCurvePrivateNumbers, numbers)
+            numbers = numbers
             data = Tlv(0x06, int2bytes(numbers.private_value, ln))
         if pin_policy:
             data += Tlv(TAG_PIN_POLICY, int2bytes(pin_policy))
